/* ---------- serial.c --------------- */

/*
 * Serial Port Communications Functions
 */

#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include "serial.h"
#include "timer.h"

int SERIAL_PORT = 1;

static char serial_buffer[SERIAL_BUFFER_SIZE];
static int XON_input_pending;
static int XON_output_pending;
char *next_serial_in  = serial_buffer;
char *next_serial_out = serial_buffer;
int xonxoff_enabled = TRUE,
    TIME_OUT        = 10,   /* number of seconds to time out  */
    buffer_count;
/* ---------------- serial port addresses ----------------- */
/* - 8250 UART base port address:  COM1 = 3f8, COM2 = 2f8 - */
int BASE, TX_DATA, RX_DATA, DIV_LSB, DIV_MSB, INT_ENABLE, INT_IDENT,
    LINE_CTL, MODEM_CTL, LINE_STATUS, MODEM_STATUS;

/* --------------- serial interrupt values ------------------ */
int IRQ, SERIAL_INT, SERIAL_IRQ;

/* ----- serial port initialization parameter byte ------ */
static union {
  struct {
    unsigned word_len  : 2;
    unsigned stop_bits : 1;
    unsigned parity    : 3;
    unsigned brk       : 1;
    unsigned divlatch  : 1;
  } serial_initial_bits;
  char serial_initial_char;
} init_serial;

static void interrupt far (*old_serial_int)(void);
static void interrupt far new_serial_int(void);

/* -------- initialize the serial port ----------- */
void init_serial_port(unsigned long int speed,
		      int parity,
		      int stopbits,
		      int wordlen)
{
  BASE=(0x3f8-((SERIAL_PORT-1)<<8));
  TX_DATA=BASE;  /* transmit data         */
  RX_DATA=BASE;  /* receive data          */
  DIV_LSB=BASE; /* baud rate divisor lsb */
  DIV_MSB=(BASE + 1); /* baud rate divisor msb */
  INT_ENABLE=(BASE + 1); /* interrupt enable      */
  INT_IDENT=(BASE + 2); /* interrupt ident'n     */
  LINE_CTL=(BASE + 3); /* line control          */
  MODEM_CTL=(BASE + 4); /* modem control         */
  LINE_STATUS=(BASE + 5); /* line status           */
  MODEM_STATUS=(BASE + 6); /* modem status          */
  IRQ=(4 - (SERIAL_PORT - 1)); /* 0-7 = IRQ0-IRQ7     */
  SERIAL_INT=(12 - (SERIAL_PORT - 1));  /* interrupt vector    */
  SERIAL_IRQ=(~(1 << IRQ));
  init_serial.serial_initial_bits.parity = parity == 2 ? 3 : parity;
  init_serial.serial_initial_bits.stop_bits = stopbits - 1;
  init_serial.serial_initial_bits.word_len = wordlen - 5;
  init_serial.serial_initial_bits.brk = 0;
  init_serial.serial_initial_bits.divlatch = 1;
  outp(LINE_CTL, init_serial.serial_initial_char);
  outp(DIV_LSB, (char) ((115200L / speed) & 255));
  outp(DIV_MSB, (char) ((115200L / speed) >> 8));
  init_serial.serial_initial_bits.divlatch = 0;
  outp(LINE_CTL, init_serial.serial_initial_char);
  /* ------ attach to the serial interrupt vector --------- */
  if (old_serial_int == NULL)
    old_serial_int = getvect(SERIAL_INT);
  setvect(SERIAL_INT, new_serial_int);
  outp(MODEM_CTL, (inp(MODEM_CTL) | DTR | RTS | OUT2));
  outp(PIC01, (inp(PIC01) & SERIAL_IRQ));
  outp(INT_ENABLE, DATA_READY);
  outp(PIC00, EOI);
  /* ----- flush existing serial interrupts ------ */
  inp(RX_DATA);
  inp(INT_IDENT);
  inp(LINE_STATUS);
  inp(MODEM_STATUS);
}

/* ------ restore the serial interrupt vector ---------- */
void restore_serial_int(void)
{
  if (old_serial_int)
    setvect(SERIAL_INT, old_serial_int);
}

/* ------- clear the serial input buffer --------- */
void clear_serial_queue(void)
{
  next_serial_in = next_serial_out = serial_buffer;
  buffer_count = 0;
}

/* ---- serial input interrupt service routine ------- */
static void interrupt far new_serial_int(void)
{
  int c;

  outp(PIC00, EOI);
  if (next_serial_in == serial_buffer + SERIAL_BUFFER_SIZE)
    next_serial_in = serial_buffer;
  c = inp(RX_DATA);
  if (xonxoff_enabled)
    if (c == XOFF)
      XON_input_pending = 1;
    else
      if (c == XON)
        XON_input_pending = 0;
  if (!xonxoff_enabled || (c != XON && c != XOFF))
    {
      *next_serial_in++ = (char) c;
      buffer_count++;
    }

  if (xonxoff_enabled && !XON_output_pending && buffer_count > THRES_HOLD)
    {
      while ((inp(LINE_STATUS) & XMIT_DATA_READY) == 0)
        ;
      outp(TX_DATA, XOFF);
      XON_output_pending = 1;
    }
}

/* ---- read a character from the input buffer ----- */
int read_serial(void)
{
/*  set_timer(TIME_OUT);
  while (!input_char_ready())
    if (timed_out())
      return FALSE; */
  if (!input_char_ready())
    return FALSE;
  if (next_serial_out == serial_buffer + SERIAL_BUFFER_SIZE)
    next_serial_out = serial_buffer;
  buffer_count--;
  if (XON_output_pending && buffer_count < SAFETY_LEVEL)
    {
      XON_output_pending = 0;
      write_serial(XON);
    }
  return *next_serial_out++;
}

/*  read a character from the input buffer for the xmodem protocol  */
int read_xserial(void)
{
  set_timer(TIME_OUT);
  while (!input_char_ready())
    if (timed_out())
      return FALSE;
  if (next_serial_out == serial_buffer + SERIAL_BUFFER_SIZE)
    next_serial_out = serial_buffer;
  buffer_count--;
  if (XON_output_pending && buffer_count < SAFETY_LEVEL)
    {
      XON_output_pending = 0;
      write_serial(XON);
    }
  return *next_serial_out++;
}


/* ---- write a character to the serial port ----- */
int write_serial(int c)
{
/*  while (XON_input_pending)
    ;
    */
  set_timer(TIME_OUT);
  while ((inp(LINE_STATUS) & XMIT_DATA_READY) == 0)
    if (timed_out())
      return FALSE;
  outp(TX_DATA, c);
  return TRUE;
}

/* ---- write a character to the serial port ----- */
int write_xserial(int c)
{
  while (XON_input_pending)
    ;
  set_timer(TIME_OUT);
  while ((inp(LINE_STATUS) & XMIT_DATA_READY) == 0)
    if (timed_out())
      return FALSE;
  outp(TX_DATA, c);
  return TRUE;
}

